/****************************************************************************
*
*                        BIOS emulator and interface
*                      to Realmode X86 Emulator Library
*
*               Copyright (C) 1996-1999 SciTech Software, Inc.
*
*  ========================================================================
*
*  Permission to use, copy, modify, distribute, and sell this software and
*  its documentation for any purpose is hereby granted without fee,
*  provided that the above copyright notice appear in all copies and that
*  both that copyright notice and this permission notice appear in
*  supporting documentation, and that the name of the authors not be used
*  in advertising or publicity pertaining to distribution of the software
*  without specific, written prior permission.  The authors makes no
*  representations about the suitability of this software for any purpose.
*  It is provided "as is" without express or implied warranty.
*
*  THE AUTHORS DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
*  INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
*  EVENT SHALL THE AUTHORS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
*  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
*  USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
*  OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
*  PERFORMANCE OF THIS SOFTWARE.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
* Developer:    Kendall Bennett
*
* Description:  Module to implement warm booting of all PCI/AGP controllers
*               on the bus. We use the x86 real mode emulator to run the
*               BIOS on the primary and secondary controllers to bring
*               the cards up.
*
****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include "biosemu.h"

/*------------------------- Global Variables ------------------------------*/

static int              NumPCI = -1;
static PCIDeviceInfo    *PCI;
static int              *BridgeIndex;
static int              *PCIBridgeIndex;
static int              *DeviceIndex;
static ulong            *BIOSImagePhys;
static int              *BIOSImageBAR;
static void             **BIOSImage;
static BE_VGAInfo       *VGAInfo;
static int              NumBridges;
static int              NumDevices;
static PCIDeviceInfo    *AGPDevice = NULL;
static PCIDeviceInfo    *AGPBridge = NULL;
static PCIDeviceInfo    *AGPHostBridge = NULL;
static u32              debugFlags = 0;
static ibool            useV86 = false;
static ibool            forcePost = false;

/* Length of the BIOS image */

#define MAX_BIOSLEN         (64 * 1024L)
#define FINAL_BIOSLEN       (32 * 1024L)

/* Macro to determine if the VGA is enabled and responding */

#define VGA_NOT_ACTIVE()    (forcePost || (PM_inpb(0x3CC) == 0xFF) || ((PM_inpb(0x3CC) & 0x2) == 0))

/*-------------------------- Implementation -------------------------------*/

/****************************************************************************
RETURNS:
The address to use to map the secondary BIOS (PCI/AGP devices)

REMARKS:
Searches all the PCI base address registers for the device looking for a
memory mapping that is large enough to hold our ROM BIOS. We usually end up
finding the framebuffer mapping (usually BAR 0x10), and we use this mapping
to map the BIOS for the device into. We use a mapping that is already
assigned to the device to ensure the memory range will be passed through
by any PCI->PCI or AGP->PCI bridge that may be present.

NOTE: Usually this function is only used for AGP devices, but it may be
      used for PCI devices that have already been POST'ed and the BIOS
      ROM base address has been zero'ed out.

NOTE: This function leaves the original memory aperture disabled by leaving
      it programmed to all 1's. It must be restored to the correct value
      later.
****************************************************************************/
static ulong PCI_findBIOSAddr(
    int device)
{
    ulong   base,size;
    int     bar;

    for (bar = 0x10; bar <= 0x14; bar++) {
        base = PCI_readPCIRegL(bar,device);
        if (!(base & 0x1)) {
            PCI_writePCIRegL(bar,0xFFFFFFFF,device);
            size = PCI_readPCIRegL(bar,device) & ~0xFF;
            size = ~size+1;
            if (size >= MAX_BIOSLEN)
                return base & ~0xFF;
            }
        }
    return 0;
}

/****************************************************************************
REMARKS:
Re-writes the PCI base address registers for the secondary PCI controller
with the values from our initial PCI bus enumeration. This fixes up the
values after we have POST'ed the secondary display controller BIOS, which
may have incorrectly re-programmed the base registers the same as the
primary display controller (the case for identical S3 cards).
{secret}
****************************************************************************/
void _PCI_fixupSecondaryBARs(void)
{
    int i;

    for (i = 0; i < NumDevices; i++) {
        PCI_writePCIRegL(0x10,PCI[DeviceIndex[i]].u.type0.BaseAddress10,i);
        PCI_writePCIRegL(0x14,PCI[DeviceIndex[i]].u.type0.BaseAddress14,i);
        PCI_writePCIRegL(0x18,PCI[DeviceIndex[i]].u.type0.BaseAddress18,i);
        PCI_writePCIRegL(0x1C,PCI[DeviceIndex[i]].u.type0.BaseAddress1C,i);
        PCI_writePCIRegL(0x20,PCI[DeviceIndex[i]].u.type0.BaseAddress20,i);
        PCI_writePCIRegL(0x24,PCI[DeviceIndex[i]].u.type0.BaseAddress24,i);
        }
}

/****************************************************************************
PARAMETERS:
length  - Size of base address to allocate

RETURNS:
Free base address region

REMARKS:
This function finds a free slot to put the PCI base address register
and programs the register to that location. This fixes up bad BIOS BAR
allocation that occurs on some graphics cards and devices.
****************************************************************************/
static ulong PCI_findFreeBAR(
    ulong length)
{
    int     i,j,valid;
    ulong   freeBase,*base,*size,minBase = 1024 * 1024L * 1024L;

    for (freeBase = 0xFFFFFFFF - length + 1; freeBase > minBase; freeBase -= length) {
        valid = true;
        for (i = 0; i < NumPCI; i++) {
            base = (ulong*)&(PCI[i].u.type0.BaseAddress10);
            size = (ulong*)&(PCI[i].u.type0.BaseAddress10Len);
            for (j = 0x10; j <= 0x24; j += 4, base++, size++) {
                if ((*base & 0x1) || *size == 0)
                    continue;
                if (*base < freeBase && (*base + *size) > freeBase) {
                    valid = false;
                    break;
                    }
                if (freeBase < *base && (freeBase + length) > *base) {
                    valid = false;
                    break;
                    }
                }
            }
        if (valid)
            return freeBase;
        }
    return 0;
}


/****************************************************************************
REMARKS:
Fixes up the PCI base address values for the display device, to compensate
for cases where the BAR's are not properly allocated by the BIOS (ie:
a zero base address).
****************************************************************************/
static void PCI_fixupBAR(
    int device,
    int bar,
    u32 *base,
    u32 *size)
{
    /* First measure the size of the BAR */
    *base = PCI_accessReg(bar,0,PCI_READ_DWORD,&PCI[device]);
    if (!(*base & 0x1)) {
        PCI_accessReg(bar,0xFFFFFFFF,PCI_WRITE_DWORD,&PCI[device]);
        *size = PCI_accessReg(bar,0,PCI_READ_DWORD,&PCI[device]) & ~0xFF;
        *size = ~*size+1;
        PCI_accessReg(bar,*base,PCI_WRITE_DWORD,&PCI[device]);
        if (*size > 0 && *base == 0) {
            *base = PCI_findFreeBAR(*size);
            PCI_accessReg(bar,*base,PCI_WRITE_DWORD,&PCI[device]);
            }
        }
}

/****************************************************************************
PARAMETERS:
info    - PCI device information to search for AGP capabilities block

RETURNS:
Offset in config space to start of AGP capabilities block, 0 if none found.
****************************************************************************/
static uchar PCI_findAGPCaps(
    PCIDeviceInfo *info,
    PCIAGPCapability *caps)
{
    uchar capPtr;

    // Check the Status register to see if this device supports capability
    // lists. If not, it is not an AGP-compliant device.
    if ((info->Status & PCI_statusCapabilitiesList) == 0)
        return 0;

    // The device supports capability lists, find the AGP capabilities
    if ((info->HeaderType & (~PCI_multiFunctionType)) == PCI_bridgeType)
        capPtr = info->u.type1.CapabilitiesPointer;
    else
        capPtr = info->u.type0.CapabilitiesPointer;

    // Search for the AGP capabilities list
    while (capPtr != 0) {
        // Read the start of the capabilities block at this offset
        PCI_readRegBlock(info,capPtr,&caps->h,sizeof(caps->h));
        if (caps->h.capsID == PCI_capsAGP) {
            // Found the AGP capabilities block
            return capPtr;
            }
        else {
            // Some other capability, so keep looking
            capPtr = caps->h.next;
            }
        }

    // If we get here, then no AGP capabilities list was found!
    return 0;
}

/****************************************************************************
PARAMETERS:
info    - PCI device information to get for AGP capabilities block for
caps    - Place to store the returned master device capabilities structure

RETURNS:
True on success, false on failure
****************************************************************************/
static ibool PCI_getAGPCaps(
    PCIDeviceInfo *info,
    PCIAGPCapability *caps)
{
    uchar capPtr;

    // Find the header block for the AGP capabilities list
    if ((capPtr = PCI_findAGPCaps(info,caps)) == 0)
        return false;

    // Now read the rest of the AGP capabilities list
    PCI_readRegBlock(info,capPtr+sizeof(caps->h),&caps->h+1,sizeof(*caps)-sizeof(caps->h));
    return true;
}

/****************************************************************************
RETURNS:
True if successfully initialised, false if not.

REMARKS:
This function executes the BIOS POST code on the controller. We assume that
at this stage the controller has its I/O and memory space enabled and
that all other controllers are in a disabled state.
****************************************************************************/
static void PCI_doBIOSPOST(
    int device,
    ulong BIOSPhysAddr,
    void *mappedBIOS,
    ulong BIOSLen)
{
    RMREGS          regs;
    RMSREGS         sregs;

    // Determine the value to store in AX for BIOS POST
    regs.x.ax = (u16)(PCI[DeviceIndex[device]].slot.i >> 8);
/*    if (useV86) {
// TODO: Restore this again!!

        // Post the BIOS using the PM functions (ie: v86 mode on Linux)
        if (!PM_doBIOSPOST(regs.x.ax,BIOSPhysAddr,mappedBIOS,BIOSLen)) {
            // If the PM function fails, this probably means are we are on
            // DOS and can't re-map the real mode 0xC0000 region. In this
            // case if the device is the primary, we can use the real
            // BIOS at 0xC0000 directly.
            if (device == 0)
                PM_doBIOSPOST(regs.x.ax,0xC0000,mappedBIOS,BIOSLen);
            }
        }
    else */{
        // Setup the X86 emulator for the VGA BIOS
        BE_setVGA(&VGAInfo[device]);

        // Execute the BIOS POST code
        BE_callRealMode(0xC000,0x0003,&regs,&sregs);

        // Cleanup and exit
        BE_getVGA(&VGAInfo[device]);
        }
}

/****************************************************************************
REMARKS:
Adjusts access to the specified device, taking into account any bridges
that the device accesses must cross.
****************************************************************************/
static void AdjustDeviceAccess(
    int deviceIndex,
    int commandMask)
{
    PCIDeviceInfo *device = &PCI[DeviceIndex[deviceIndex]];

    PCI_accessReg(0x4,(device->Command & ~0x7) | commandMask,PCI_WRITE_WORD,device);
    if (device->slot.p.Bus > 0) {
        // Enable or disable the PCI->PCI bridge VGA access control for this device
        // depending on whether we are enabling I/O access for this device. Only
        // one device can have the VGA I/O accesses routed to it over the PCI bridge.
        PCIDeviceInfo *bridge = &PCI[PCIBridgeIndex[device->slot.p.Bus]];
        if (commandMask & 0x1)
            PCI_accessReg(0x3E,bridge->u.type1.BridgeControl | 0x8,PCI_WRITE_WORD,bridge);
        else
            PCI_accessReg(0x3E,bridge->u.type1.BridgeControl & ~0x8,PCI_WRITE_WORD,bridge);
        }
}

/****************************************************************************
REMARKS:
Disables the specified device, taking into account any bridges that the
device accesses must cross.
****************************************************************************/
static void DisableDevice(
    int deviceIndex)
{
    PCIDeviceInfo *device = &PCI[DeviceIndex[deviceIndex]];

    PCI_accessReg(0x4,0,PCI_WRITE_WORD,device);
    if (device->slot.p.Bus > 0) {
        // Disable the PCI->PCI bridge VGA access control for this device
        PCIDeviceInfo *bridge = &PCI[PCIBridgeIndex[device->slot.p.Bus]];
        PCI_accessReg(0x3E,bridge->u.type1.BridgeControl & ~0x8,PCI_WRITE_WORD,bridge);
        }
}

/****************************************************************************
RETURNS:
True if successfully initialised, false if not.

REMARKS:
Loads and POST's the secondary controllers BIOS, directly from the BIOS
image we can extract over the PCI bus.
****************************************************************************/
static ibool PCI_postControllers(void)
{
    int     device;
    ulong   BIOSImageLen,mappedBIOSPhys;
    uchar   *mappedBIOS,*copyOfBIOS;
    char    filename[PM_MAX_PATH];
    FILE    *f;

    // Disable the primary display controller
    DisableDevice(0);

    // Now POST all the secondary controllers
    for (device = 0; device < NumDevices; device++) {
        // Skip the device if it is not enabled (probably an ISA device)
        if (DeviceIndex[device] == -1)
            continue;

        // Enable secondary display controller.
        AdjustDeviceAccess(device,0x7);

        // Check if the controller has already been POST'ed
        if (VGA_NOT_ACTIVE()) {
            // Find a viable place to map the secondary PCI BIOS image and map it
            printk("Device %d not enabled, so attempting warm boot it\n", device);

            // For AGP devices (and PCI devices that do have the ROM base
            // address zero'ed out) we have to map the BIOS to a location
            // that is passed by the AGP bridge to the bus. Some AGP devices
            // have the ROM base address already set up for us, and some
            // do not (we map to one of the existing BAR locations in
            // this case).
            mappedBIOS = NULL;
            if (PCI[DeviceIndex[device]].u.type0.ROMBaseAddress != 0)
                mappedBIOSPhys = PCI[DeviceIndex[device]].u.type0.ROMBaseAddress & ~0xF;
            else
                mappedBIOSPhys = PCI_findBIOSAddr(device);
            printk("Mapping BIOS image to 0x%08X\n", mappedBIOSPhys);
            mappedBIOS = PM_mapPhysicalAddr(mappedBIOSPhys,MAX_BIOSLEN-1,false);
            PCI_writePCIRegL(0x30,mappedBIOSPhys | 0x1,device);
            BIOSImageLen = mappedBIOS[2] * 512;
            if ((copyOfBIOS = malloc(BIOSImageLen)) == NULL)
                return false;
            memcpy(copyOfBIOS,mappedBIOS,BIOSImageLen);
            PM_freePhysicalAddr(mappedBIOS,MAX_BIOSLEN-1);

            // Allocate memory to store copy of BIOS from secondary controllers
            VGAInfo[device].pciInfo = &PCI[DeviceIndex[device]];
            VGAInfo[device].BIOSImage = copyOfBIOS;
            VGAInfo[device].BIOSImageLen = BIOSImageLen;

            // Restore device mappings
            PCI_writePCIRegL(0x30,PCI[DeviceIndex[device]].u.type0.ROMBaseAddress,device);
            PCI_writePCIRegL(0x10,PCI[DeviceIndex[device]].u.type0.BaseAddress10,device);
            PCI_writePCIRegL(0x14,PCI[DeviceIndex[device]].u.type0.BaseAddress14,device);

            // Now execute the BIOS POST for the device
            if (copyOfBIOS[0] == 0x55 && copyOfBIOS[1] == 0xAA) {
                printk("Executing BIOS POST for controller.\n");
                PCI_doBIOSPOST(device,mappedBIOSPhys,copyOfBIOS,BIOSImageLen);
                }

            // Reset the size of the BIOS image to the final size
            VGAInfo[device].BIOSImageLen = FINAL_BIOSLEN;

            // Save the BIOS and interrupt vector information to disk
            sprintf(filename,"%s/bios.%02d",PM_getSNAPConfigPath(),device);
            if ((f = fopen(filename,"wb")) != NULL) {
                fwrite(copyOfBIOS,1,FINAL_BIOSLEN,f);
                fwrite(VGAInfo[device].LowMem,1,sizeof(VGAInfo[device].LowMem),f);
                fclose(f);
                }
            }
        else {
            // Allocate memory to store copy of BIOS from secondary controllers
            if ((copyOfBIOS = malloc(FINAL_BIOSLEN)) == NULL)
                return false;
            VGAInfo[device].pciInfo = &PCI[DeviceIndex[device]];
            VGAInfo[device].BIOSImage = copyOfBIOS;
            VGAInfo[device].BIOSImageLen = FINAL_BIOSLEN;

            // Load the BIOS and interrupt vector information from disk
            sprintf(filename,"%s/bios.%02d",PM_getSNAPConfigPath(),device);
            if ((f = fopen(filename,"rb")) != NULL) {
                fread(copyOfBIOS,1,FINAL_BIOSLEN,f);
                fread(VGAInfo[device].LowMem,1,sizeof(VGAInfo[device].LowMem),f);
                fclose(f);
                }
            }

        // Fix up all the secondary PCI base address registers
        // (restores them all from the values we read previously)
        _PCI_fixupSecondaryBARs();

        // Disable the secondary controller
        DisableDevice(device);
        }

    // Reenable primary display controller
    AdjustDeviceAccess(0,0x7);

    // Free physical BIOS image mapping
    PM_freePhysicalAddr(mappedBIOS,MAX_BIOSLEN-1);

    // Restore the X86 emulator BIOS info to primary controller
    if (!useV86)
        BE_setVGA(&VGAInfo[0]);
    return true;
}

/****************************************************************************
REMARKS:
Enumerates the device index information to the end of the config info line.
****************************************************************************/
static void LogDeviceIndex(
    int i)
{
    int index;

    for (index = 0; index < NumDevices; index++) {
        if (DeviceIndex[index] == i)
            break;
        }
    if (index < NumDevices) {
        if (&PCI[i] == AGPDevice)
            printk("<- %d (AGP Card)\n", index);
        else
            printk("<- %d\n", index);
        }
    else
        printk("\n");
}

/****************************************************************************
REMARKS:
Enumerates the PCI bus and dumps the PCI configuration information to the
log file.
****************************************************************************/
static void EnumeratePCI(void)
{
    int             i,index;
    PCIDeviceInfo   *info;

    printk("Displaying enumeration of PCI bus (%d devices, %d display devices)\n",
        NumPCI, NumDevices);
    for (index = 0; index < NumDevices; index++)
        printk("  Display device %d is PCI device %d\n",index,DeviceIndex[index]);
    printk("\n");
    printk("Bus Slot Fnc DeviceID  SubSystem Rev Class IRQ Base10h  Base14h  Cmd\n");
    for (i = 0; i < NumPCI; i++) {
        printk("%2d   %2d  %2d  %04X:%04X %04X:%04X %02X  %02X:%02X %02X  %08X %08X %04X   ",
            PCI[i].slot.p.Bus,
            PCI[i].slot.p.Device,
            PCI[i].slot.p.Function,
            PCI[i].VendorID,
            PCI[i].DeviceID,
            PCI[i].u.type0.SubSystemVendorID,
            PCI[i].u.type0.SubSystemID,
            PCI[i].RevID,
            PCI[i].BaseClass,
            PCI[i].SubClass,
            PCI[i].u.type0.InterruptLine,
            PCI[i].u.type0.BaseAddress10,
            PCI[i].u.type0.BaseAddress14,
            PCI[i].Command);
        LogDeviceIndex(i);
        }
    printk("\n");
    printk("DeviceID  Stat Ifc Cch Lat Hdr BIST Int Base18h  Base1Ch  Base20h  Base24h  \n");
    for (i = 0; i < NumPCI; i++) {
        printk("%04X:%04X %04X  %02X  %02X  %02X  %02X  %02X  %02X  %08X %08X %08X %08X ",
            PCI[i].VendorID,
            PCI[i].DeviceID,
            PCI[i].Status,
            PCI[i].Interface,
            PCI[i].CacheLineSize,
            PCI[i].LatencyTimer,
            PCI[i].HeaderType,
            PCI[i].BIST,
            PCI[i].u.type0.InterruptPin,
            PCI[i].u.type0.BaseAddress18,
            PCI[i].u.type0.BaseAddress1C,
            PCI[i].u.type0.BaseAddress20,
            PCI[i].u.type0.BaseAddress24);
        LogDeviceIndex(i);
        }
    printk("\n");
    printk("DeviceID  BAR10Len BAR14Len BAR18Len BAR1CLen BAR20Len BAR24Len ROMLen\n");
    for (i = 0; i < NumPCI; i++) {
        printk("%04X:%04X %08X %08X %08X %08X %08X %08X %08X ",
            PCI[i].VendorID,
            PCI[i].DeviceID,
            PCI[i].u.type0.BaseAddress10Len,
            PCI[i].u.type0.BaseAddress14Len,
            PCI[i].u.type0.BaseAddress18Len,
            PCI[i].u.type0.BaseAddress1CLen,
            PCI[i].u.type0.BaseAddress20Len,
            PCI[i].u.type0.BaseAddress24Len,
            PCI[i].u.type0.ROMBaseAddressLen);
        LogDeviceIndex(i);
        }
    printk("\n");
    printk("Displaying enumeration of %d bridge devices\n",NumBridges);
    printk("\n");
    printk("DeviceID  P# S# B# IOB  IOL  MemBase  MemLimit PreBase  PreLimit Ctrl\n");
    for (i = 0; i < NumBridges; i++) {
        info = &PCI[BridgeIndex[i]];
        printk("%04X:%04X %02X %02X %02X %04X %04X %08X %08X %08X %08X %04X ",
            info->VendorID,
            info->DeviceID,
            info->u.type1.PrimaryBusNumber,
            info->u.type1.SecondayBusNumber,
            info->u.type1.SubordinateBusNumber,
            ((u16)info->u.type1.IOBase << 8) & 0xF000,
            info->u.type1.IOLimit ?
                ((u16)info->u.type1.IOLimit << 8) | 0xFFF : 0,
            ((u32)info->u.type1.MemoryBase << 16) & 0xFFF00000,
            info->u.type1.MemoryLimit ?
                ((u32)info->u.type1.MemoryLimit << 16) | 0xFFFFF : 0,
            ((u32)info->u.type1.PrefetchableMemoryBase << 16) & 0xFFF00000,
            info->u.type1.PrefetchableMemoryLimit ?
                ((u32)info->u.type1.PrefetchableMemoryLimit << 16) | 0xFFFFF : 0,
            info->u.type1.BridgeControl);
        if (info == AGPBridge)
            printk("<- AGP Bridge\n");
        else if (info == AGPHostBridge)
            printk("<- AGP Host Bridge\n");
        else
            printk("\n");
        }
    printk("\n");
}

/****************************************************************************
RETURNS:
Number of display devices found.

REMARKS:
This function enumerates the number of available display devices on the
PCI bus, and returns the number found.
****************************************************************************/
static int PCI_enumerateDevices(void)
{
    int                 i,j,size,maxDevices;
    PCIDeviceInfo       *info;
    PCIAGPCapability    agpCaps;

    // If this is the first time we have been called, enumerate all
    // devices on the PCI bus.
    if (NumPCI == -1) {
        // First determine how many devices are on the PCI bus, and allocate
        // enough memory to describe all the devices in one memory block.
        if ((NumPCI = PCI_getNumDevices()) == 0)
            return -1;
        size = NumPCI * (sizeof(PCIDeviceInfo) + sizeof(int) + sizeof(int) + sizeof(int));
        if ((PCI = calloc(1,size)) == NULL)
            return -1;
        BridgeIndex = (int*)(PCI + NumPCI);
        PCIBridgeIndex = (int*)(BridgeIndex + NumPCI);
        DeviceIndex = (int*)(PCIBridgeIndex + NumPCI);
        for (i = 0; i < NumPCI; i++)
            PCI[i].dwSize = sizeof(PCI[i]);
        if (PCI_enumerate(PCI) == 0)
            return -1;

        // Build a list of all PCI bridge devices
        for (i = 0,NumBridges = 0,BridgeIndex[0] = -1; i < NumPCI; i++) {
            if (PCI[i].BaseClass == PCI_BRIDGE_CLASS) {
                BridgeIndex[NumBridges++] = i;
                if (PCI[i].SubClass == PCI_HOST_BRIDGE_SUBCLASS && !AGPHostBridge) {
                    if (PCI_getAGPCaps(&PCI[i],&agpCaps))
                        AGPHostBridge = &PCI[i];
                    }
                if (PCI[i].SubClass == PCI_PCI_BRIDGE_SUBCLASS) {
                    // Keep an index of all PCI->PCI bridges based on their secondary
                    // bus number. This allows us to quickly access to bridge control
                    // registers for a device based on the secondary bus number.
                    PCIBridgeIndex[PCI[i].u.type1.SecondayBusNumber] = i;
                    }
                }
            }

        // Now build a list of all display class devices
        DeviceIndex[0] = -1;
        for (i = 0,NumDevices = 1,DeviceIndex[0] = -1; i < NumPCI; i++) {
            if (PCI_IS_DISPLAY_CLASS(&PCI[i])) {
                PCI_fixupBAR(i,0x10,&PCI[i].u.type0.BaseAddress10,&PCI[i].u.type0.BaseAddress10Len);
                PCI_fixupBAR(i,0x14,&PCI[i].u.type0.BaseAddress14,&PCI[i].u.type0.BaseAddress14Len);
                if ((PCI[i].Command & 0x3) == 0x3)
                    DeviceIndex[0] = i;
                else
                    DeviceIndex[NumDevices++] = i;
                if (PCI[i].slot.p.Bus != 0 && PCI_getAGPCaps(&PCI[i],&agpCaps)) {
                    // This device is on a different bus than the primary
                    // PCI bus, so it is probably an AGP device. Find the
                    // AGP bus device that controls that bus so we can
                    // control it.
                    for (j = 0; j < NumBridges; j++) {
                        info = &PCI[BridgeIndex[j]];
                        if (info->u.type1.SecondayBusNumber == (int)PCI[i].slot.p.Bus) {
                            AGPDevice = &PCI[i];
                            AGPBridge = info;
                            break;
                            }
                        }
                    }
                }
            }

        // If we did not find the primary PCI device, then we must have a
        // VLB/ISA device.
        if (DeviceIndex[0] == -1)
            NumDevices = 0;

        // Now display the list of PCI devices
        EnumeratePCI();

        // Allocate memory for our array of objects in a single memory block
        maxDevices = (NumDevices == 0) ? 1 : NumDevices;
        size = maxDevices * (sizeof(ulong) + sizeof(int) + sizeof(void*) + sizeof(BE_VGAInfo));
        if ((BIOSImagePhys = calloc(1,size)) == NULL)
            return -1;
        BIOSImageBAR = (int*)(BIOSImagePhys + maxDevices);
        BIOSImage = (void**)(BIOSImageBAR + maxDevices);
        VGAInfo = (BE_VGAInfo*)(BIOSImage + maxDevices);

        // Map the BIOS image for the primary controller at 0xC0000
        BIOSImagePhys[0] = 0xC0000;
        BIOSImage[0] = PM_mapPhysicalAddr(0xC0000,MAX_BIOSLEN-1,true);
        }
    return NumDevices;
}

FILE *logfile;

void printk(const char *fmt, ...)
{
    va_list argptr;
    va_start(argptr, fmt);
    vfprintf(logfile, fmt, argptr);
    fflush(logfile);
    va_end(argptr);
}

int main(int argc,char *argv[])
{
    while (argc > 1) {
        if (stricmp(argv[1],"-usev86") == 0) {
            useV86 = true;
            }
        else if (stricmp(argv[1],"-force") == 0) {
            forcePost = true;
            }
#ifdef  DEBUG
        else if (stricmp(argv[1],"-decode") == 0) {
            debugFlags |= DEBUG_DECODE_F;
            }
        else if (stricmp(argv[1],"-iotrace") == 0) {
            debugFlags |= DEBUG_IO_TRACE_F;
            }
#endif
        else {
            printf("Usage: warmboot [-usev86] [-force] [-decode] [-iotrace]\n");
            exit(-1);
            }
        argc--;
        argv++;
        }
    if ((logfile = fopen("warmboot.log","w")) == NULL)
        exit(1);

    // Enumerate all devices (which POST's them at the same time)
    PM_init();
    if (PCI_enumerateDevices() < 1) {
        printk("No PCI display devices found!\n");
        return -1;
        }

    if (!useV86) {
        // Initialise the x86 BIOS emulator
        BE_init(debugFlags,65536,&VGAInfo[0]);
        }

    // Post all the display controller BIOS'es
    PCI_postControllers();

    // Cleanup and exit the emulator
    if (!useV86)
        BE_exit();
    fclose(logfile);
    return 0;
}

